/*
 * devglue_core.c
 *
 * Copyright (c) 2016 Socionext Inc.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2  of
 * the License as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#define pr_fmt(fmt) "devglue: " fmt

#include <linux/cdev.h>			// cdev_*
#include <linux/clk.h>			// clk
#include <linux/fs.h>			// chrdev
#include <linux/module.h>
#include <linux/of_irq.h>		// of_*
#include <linux/of_platform.h>		// of_*
#include <linux/platform_device.h>	// platform_*
#include <linux/slab.h>			// kzalloc

#include "devglue.h"

#define TRUE    (1UL)
#define FALSE   (0UL)

struct devglue_regs {
	void __iomem *vbase;
	phys_addr_t  pbase;
	size_t       size;
	char         name[DEVGLUE_MAX_NAMELEN];
};

struct devglue_irqs {
	unsigned int irq;
	unsigned int hwirq;
	char         name[DEVGLUE_MAX_NAMELEN];
	int status;
};

struct devglue_clks {
	struct clk   *clk;
	char         name[DEVGLUE_MAX_NAMELEN];
};

struct devglue_uniphier {
	struct device *dev;
	struct class *class;
	int nr_regs;
	struct devglue_regs *regs;
	int nr_irqs;
	struct devglue_irqs *irqs;
	int nr_clks;
	struct devglue_clks *clks;
};

struct devglue_subdata {
	int n_dev;
	dev_t id;
	struct cdev cdev;
	struct device *dev[DEVGLUE_MAX_DEVNUM];
};

/* FIXME: external module can't specify this driver, then keep the work pointer inside. */
static struct devglue_uniphier *_devglue = NULL;

/*
 * initialize
 */
static void devglue_init(struct devglue_uniphier *devglue)
{
	return;
}

static void devglue_exit(struct devglue_uniphier *devglue)
{
	return;
}

/*
 * helper functions
 */
static int _devglue_search_reg_byname(struct devglue_uniphier *devglue, const char *name)
{
	int i, index = -1;

	if (devglue == NULL)
		return -ENODEV;

	for(i=0; i<devglue->nr_regs; i++) {
		if (!strncmp(devglue->regs[i].name, name, DEVGLUE_MAX_NAMELEN)) {
			index = i;
			break;
		}
	}
	if (index < 0)
		return -EINVAL;

	return index;
}

void __iomem *devglue_ioremap(const char *name, enum devglue_memory_type type)
{
	struct devglue_uniphier *devglue = _devglue;
	void __iomem *base;
	int index = -1;

	if (devglue == NULL)
		return ERR_PTR(-ENODEV);

	index = _devglue_search_reg_byname(devglue, name);
	if (index < 0) {
		dev_err(devglue->dev, "Reg \"%s\" not found\n", name);
		return ERR_PTR(-EINVAL);
	}

	switch(type) {
	case DEVGLUE_MEMORY_WC:
		base = devm_ioremap_wc(devglue->dev,
				     devglue->regs[index].pbase, devglue->regs[index].size);
		break;
	case DEVGLUE_MEMORY_UC:
		base = devm_ioremap_nocache(devglue->dev,
				     devglue->regs[index].pbase, devglue->regs[index].size);
		break;
	default:
		dev_err(devglue->dev, "Invalid memory type\n");
		return ERR_PTR(-EINVAL);
	}
	if (IS_ERR(base)) {
		dev_err(devglue->dev, "Failed to map resource\n");
		return base;
	}
	devglue->regs[index].vbase = base;

	return base;
}
EXPORT_SYMBOL(devglue_ioremap);

void devglue_iounmap(void __iomem *vbase)
{
	struct devglue_uniphier *devglue = _devglue;

	if (devglue == NULL)
		return;

	devm_iounmap(devglue->dev, vbase);

	return;
}
EXPORT_SYMBOL(devglue_iounmap);

int devglue_reg_get_byname(const char *name, phys_addr_t *pbase, size_t *size)
{
	struct devglue_uniphier *devglue = _devglue;
	int index = -1;

	if (devglue == NULL)
		return -ENODEV;

	if (name == NULL)
		return -EINVAL;

	index = _devglue_search_reg_byname(devglue, name);
	if (index < 0) {
		dev_err(devglue->dev, "Reg \"%s\" not found\n", name);
		return -EINVAL;
	}

	*pbase = devglue->regs[index].pbase;
	*size = devglue->regs[index].size;

	return 0;
}
EXPORT_SYMBOL(devglue_reg_get_byname);

static int _devglue_search_irq(struct devglue_uniphier *devglue, unsigned int irq)
{
	int i, index = -1;

	if (devglue == NULL)
		return -ENODEV;

	for(i=0; i<devglue->nr_irqs; i++) {
		if (devglue->irqs[i].hwirq == irq) {
			index = i;
			break;
		}
	}
	if (index < 0)
		return -EINVAL;

	return index;
}

static int _devglue_search_irq_byname(struct devglue_uniphier *devglue, const char *name)
{
	int i, index = -1;

	if (devglue == NULL)
		return -ENODEV;

	for(i=0; i<devglue->nr_irqs; i++) {
		if (!strncmp(devglue->irqs[i].name, name, DEVGLUE_MAX_NAMELEN)) {
			index = i;
			break;
		}
	}
	if (index < 0)
		return -EINVAL;

	return index;
}

int devglue_request_irq(unsigned int irq, irq_handler_t handler, unsigned long flags, const char *name, void *dev_id)
{
	struct devglue_uniphier *devglue = _devglue;
	int ret, index;

	if (devglue == NULL)
		return -ENODEV;

	index = _devglue_search_irq(devglue, irq);
	if (index < 0) {
		dev_err(devglue->dev, "Irq %d not found\n", irq);
		return -EINVAL;
	}

	ret = devm_request_irq(devglue->dev, devglue->irqs[index].irq, handler, flags, name, dev_id);
	if (ret < 0) {
		dev_err(devglue->dev, "Failed to request irq\n");
		return ret;
	}

	devglue->irqs[index].status = TRUE;
	return 0;
}
EXPORT_SYMBOL(devglue_request_irq);

int devglue_request_irq_byname(unsigned int *irqp, irq_handler_t handler, unsigned long flags, const char *name, void *dev_id)
{
	struct devglue_uniphier *devglue = _devglue;
	int ret, index = -1;

	if (devglue == NULL)
		return -ENODEV;

	if (name == NULL)
		return -EINVAL;

	index = _devglue_search_irq_byname(devglue, name);
	if (index < 0) {
		dev_err(devglue->dev, "Irq \"%s\" not found\n", name);
		return -EINVAL;
	}

	ret = devm_request_irq(devglue->dev, devglue->irqs[index].irq, handler, flags, name, dev_id);
	if (ret < 0) {
		dev_err(devglue->dev, "Failed to request irq\n");
		return ret;
	}

	devglue->irqs[index].status = TRUE;
	*irqp = devglue->irqs[index].hwirq;
	return 0;
}
EXPORT_SYMBOL(devglue_request_irq_byname);

void devglue_free_irq(unsigned int irq, void *dev_id)
{
	struct devglue_uniphier *devglue = _devglue;
	int index;

	if (devglue == NULL)
		return;

	index = _devglue_search_irq(devglue, irq);
	if (index < 0)
		return;

	if (devglue->irqs[index].status == TRUE)
		devm_free_irq(devglue->dev, devglue->irqs[index].irq, dev_id);

	devglue->irqs[index].status = FALSE;
	return;
}
EXPORT_SYMBOL(devglue_free_irq);

void devglue_enable_irq(unsigned int irq)
{
	struct devglue_uniphier *devglue = _devglue;
	int index;

	if (devglue == NULL)
		return;

	index = _devglue_search_irq(devglue, irq);
	if (index < 0)
		return;

	if (devglue->irqs[index].status == TRUE)
		enable_irq(devglue->irqs[index].irq);
	else
		irq_modify_status(devglue->irqs[index].irq, IRQ_NOAUTOEN, 0);
	return;
}
EXPORT_SYMBOL(devglue_enable_irq);

void devglue_enable_irq_byname(const char *name)
{
	struct devglue_uniphier *devglue = _devglue;
	int index;

	if (devglue == NULL)
		return;

	if (name == NULL)
		return;

	index = _devglue_search_irq_byname(devglue, name);
	if (index < 0)
		return;

	if (devglue->irqs[index].status == TRUE)
		enable_irq(devglue->irqs[index].irq);
	else
		irq_modify_status(devglue->irqs[index].irq, IRQ_NOAUTOEN, 0);
	return;
}
EXPORT_SYMBOL(devglue_enable_irq_byname);

void devglue_disable_irq(unsigned int irq)
{
	struct devglue_uniphier *devglue = _devglue;
	int index;

	if (devglue == NULL)
		return;

	index = _devglue_search_irq(devglue, irq);
	if (index < 0)
		return;

	if (devglue->irqs[index].status == TRUE)
		disable_irq(devglue->irqs[index].irq);
	else
		irq_modify_status(devglue->irqs[index].irq, 0, IRQ_NOAUTOEN);
	return;
}
EXPORT_SYMBOL(devglue_disable_irq);

void devglue_disable_irq_byname(const char *name)
{
	struct devglue_uniphier *devglue = _devglue;
	int index;

	if (devglue == NULL)
		return;

	if (name == NULL)
		return;

	index = _devglue_search_irq_byname(devglue, name);
	if (index < 0)
		return;

	if (devglue->irqs[index].status == TRUE)
		disable_irq(devglue->irqs[index].irq);
	else
		irq_modify_status(devglue->irqs[index].irq, 0, IRQ_NOAUTOEN);
	return;
}
EXPORT_SYMBOL(devglue_disable_irq_byname);

int devglue_irq_get(unsigned int irq)
{
	struct devglue_uniphier *devglue = _devglue;
	int index;

	if (devglue == NULL)
		return -ENODEV;

	index = _devglue_search_irq(devglue, irq);
	if (index < 0) {
		dev_err(devglue->dev, "Irq %d not found\n", irq);
		return -EINVAL;
	}

	return devglue->irqs[index].irq;
}
EXPORT_SYMBOL(devglue_irq_get);

int devglue_irq_get_byname(const char *name)
{
	struct devglue_uniphier *devglue = _devglue;
	int index = -1;

	if (devglue == NULL)
		return -ENODEV;

	if (name == NULL)
		return -EINVAL;

	index = _devglue_search_irq_byname(devglue, name);
	if (index < 0) {
		dev_err(devglue->dev, "Irq \"%s\" not found\n", name);
		return -EINVAL;
	}

	return devglue->irqs[index].irq;
}
EXPORT_SYMBOL(devglue_irq_get_byname);

struct device **devglue_alloc_device(const char *devpath, unsigned int num, unsigned int devno,
				     const struct file_operations *fops)
{
	struct devglue_uniphier *devglue = _devglue;
	struct devglue_subdata *dgs;
	int i, j, ret;
	char devname[DEVGLUE_MAX_PATHLEN], *p;

	if (devglue == NULL)
		return ERR_PTR(-ENODEV);

	if ((num > DEVGLUE_MAX_DEVNUM) || (devpath == NULL))
		return ERR_PTR(-EINVAL);

	/* allocate subdata work area of device */
	dgs = kzalloc(sizeof(struct devglue_subdata), GFP_KERNEL);
	if (dgs == NULL) {
		pr_err("devglue: Failed to alloc memory.\n");
		return ERR_PTR(-ENOMEM);
	}

	/* convert device file path to device name */
	memset(devname, 0, sizeof(devname));
	strncpy(devname, devpath, DEVGLUE_MAX_PATHLEN - 1);
	for(p = devname; *p; p++) {
		if (*p == '/')
			*p = '_';
	}

	/* Allocate region of character device */
	if (devno == 0) {
		if ((ret = alloc_chrdev_region(&dgs->id, 0, num, devname)) < 0) {
			pr_err("devglue: Failed to exec alloc_chrdev_region().\n");
			goto err_ret;
		}
	}
	else {
		dgs->id = MKDEV(devno, 0);
		if ((ret = register_chrdev_region(dgs->id, 1, devname)) < 0) {
			pr_err("devglue: Failed to exec register_chrdev_region().\n");
			goto err_ret;
		}
	}

	/* Initialize character device */
	cdev_init(&dgs->cdev, fops);
	if ((ret = cdev_add(&dgs->cdev, dgs->id, num)) < 0) {
		pr_err("devglue: Failed to exec cdev_add().\n");
		goto err_region;
	}

	/* Create device */
	for(i=0; i<num; i++) {
		struct device *dev;

		if (num == 1) {
			dev = device_create(devglue->class, NULL,
					    MKDEV(MAJOR(dgs->id), MINOR(dgs->id) + i),
					    NULL, "%s", devpath);
		}
		else {
			dev = device_create(devglue->class, NULL,
					    MKDEV(MAJOR(dgs->id), MINOR(dgs->id) + i),
					    NULL, "%s%d", devpath, i);
		}
		if (IS_ERR(dev)) {
			pr_err("devglue: Failed to exec device_create().\n");
			ret = PTR_ERR(dev);
			goto err_dev;
		}
		dgs->dev[i] = dev;
		dev_set_drvdata(dgs->dev[i], dgs);
		if ((devglue)&&(devglue->dev))
			of_dma_configure(dgs->dev[i], devglue->dev->of_node);
	}
	dgs->n_dev = num;

	return dgs->dev;

err_dev:
	for(j=0; j<i; j++) {
		device_destroy(devglue->class,
			       MKDEV(MAJOR(dgs->id), MINOR(dgs->id) + i));
	}
	cdev_del(&dgs->cdev);
err_region:
	unregister_chrdev_region(dgs->id, dgs->n_dev);
err_ret:
	kfree(dgs);
	return ERR_PTR(ret);
}
EXPORT_SYMBOL(devglue_alloc_device);

void devglue_free_device(struct device **dev)
{
	struct devglue_uniphier *devglue = _devglue;
	struct devglue_subdata *dgs;
	int i;

	if ((dev == NULL)||(*dev == NULL))
		return;

	dgs = dev_get_drvdata(*dev);

	for(i=0; i<dgs->n_dev; i++) {
		device_destroy(devglue->class,
			       MKDEV(MAJOR(dgs->id), MINOR(dgs->id) + i));
	}

	cdev_del(&dgs->cdev);
	unregister_chrdev_region(dgs->id, dgs->n_dev);
	kfree(dgs);

	return;
}
EXPORT_SYMBOL(devglue_free_device);

/*
 * probe driver
 */
static int devglue_probe(struct platform_device *pdev)
{
	struct resource *res;
	struct device *dev = &pdev->dev;
	struct devglue_uniphier *devglue;
	struct clk *clk;
	char *buf;
	unsigned int irq;
	int i, nr, ret = 0;
	struct of_phandle_args oirq;

	devglue = devm_kzalloc(dev, sizeof(*devglue), GFP_KERNEL);
	if (!devglue) {
		dev_err(dev, "Failed to alloc memory\n");
		return -ENOMEM;
	}
	platform_set_drvdata(pdev, devglue);
	devglue->dev = dev;

	/* map register */
	nr = 0;
	while (platform_get_resource(pdev, IORESOURCE_MEM, nr) != NULL)
		nr++;

	devglue->nr_regs = nr;
	devglue->regs = devm_kzalloc(dev, sizeof(struct devglue_regs) * nr, GFP_KERNEL);
	if (!devglue->regs) {
		dev_err(dev, "Failed to alloc memory for regs\n");
		ret = -ENOMEM;
		goto err_work;
	}

	for(i=0; i<nr; i++) {
		res = platform_get_resource(pdev, IORESOURCE_MEM, i);
		if (!res) {
			dev_err(dev, "Failed to get reg[%d]\n", i);
			ret = -ENODEV;
			goto err_regs;
		}
		ret = of_property_read_string_index(dev->of_node, "reg-names", i, (const char**)&buf);
		if (ret < 0) {
			if (ret == -ENODATA) {
				buf = NULL;
			}
			else {
				dev_err(dev, "Invalid reg name[%d]\n", i);
				goto err_regs;
			}
		}
		else {
			strncpy(devglue->regs[i].name, buf, DEVGLUE_MAX_NAMELEN);
		}
		devglue->regs[i].vbase = NULL;
		devglue->regs[i].pbase = res->start;
		devglue->regs[i].size  = resource_size(res);
		dev_dbg(dev, "registered reg[%d] \"%s\"\n", i, (buf) ? buf : "");
	}

	/* prepare interrupt */
	nr = 0;
	while ((ret = platform_get_irq(pdev, nr)) >= 0)
		nr++;

	devglue->nr_irqs = nr;
	devglue->irqs = devm_kzalloc(dev, sizeof(struct devglue_irqs) * nr, GFP_KERNEL);
	if (!devglue->irqs) {
		dev_err(dev, "Failed to alloc memory for irqs\n");
		return -ENOMEM;
	}

	for(i=0; i<nr; i++) {
		irq = platform_get_irq(pdev, i);
		if (!irq) {
			dev_err(dev, "Failed to get irq[%d]\n", i);
			ret = -ENODEV;
			goto err_irqs;
		}
		ret = of_property_read_string_index(dev->of_node, "interrupt-names", i, (const char**)&buf);
		if (ret < 0) {
			if (ret == -ENODATA) {
				buf = NULL;
			}
			else {
				dev_err(dev, "Invalid the name of irq[%d] %d\n", i,ret);
				goto err_irqs;
			}
		}
		else {
			strncpy(devglue->irqs[i].name, buf, DEVGLUE_MAX_NAMELEN);
		}
		ret = of_irq_parse_one(dev->of_node, i, &oirq);
		if (ret < 0) {
			dev_err(dev, "Invalid the number of irq[%d]\n", i);
			goto err_irqs;
		}
		devglue->irqs[i].irq = irq;
		devglue->irqs[i].hwirq = (oirq.args_count == 1) ? oirq.args[0] : oirq.args[1];
		dev_dbg(dev, "registered irq[%d] %d \"%s\"\n", i, devglue->irqs[i].hwirq, (buf) ? buf : "");
	}

	/* enable clock */
	nr = 0;
	while (1) {
		struct clk *clk = of_clk_get(dev->of_node, nr);
		if (IS_ERR(clk))
			break;
		nr++;
	}

	devglue->nr_clks = nr;
	devglue->clks = devm_kzalloc(dev, sizeof(struct devglue_clks) * nr, GFP_KERNEL);
	if (!devglue->clks) {
		dev_err(dev, "Failed to alloc memory for clks\n");
		ret = -ENOMEM;
		goto err_irqs;
	}

	for(i=0; i<nr; i++) {
		clk = of_clk_get(dev->of_node, i);
		if (!clk) {
			dev_err(dev, "Failed to get clk[%d]\n", i);
			ret = -ENODEV;
			goto err_clks;
		}
		ret = clk_prepare_enable(clk);
		if (ret) {
			dev_err(dev, "Failed to register clk[%d]\n", i);
			goto err_clks;
		}
		ret = of_property_read_string_index(dev->of_node, "clock-names", i, (const char**)&buf);
		if (ret < 0) {
			dev_err(dev, "Invalid the name of clk[%d]\n", i);
			goto err_clks;
		}
		devglue->clks[i].clk  = clk;
		strncpy(devglue->clks[i].name, buf, DEVGLUE_MAX_NAMELEN);
	}

	/* create device class */
	devglue->class = class_create(THIS_MODULE, "devglue");
	if (IS_ERR(devglue->class)) {
		dev_err(dev, "Failed to create 'devglue' class\n");
		ret = PTR_ERR(devglue->class);
		goto err_clks;
	}

	/* initialize driver */
	devglue_init(devglue);

	/* populate driver */
	ret = of_platform_populate(dev->of_node, NULL, NULL, dev);
	if (ret) {
		dev_err(dev, "Failed to register driver(%d)\n", ret);
		goto err_exit;
	}

	_devglue = devglue;
	return 0;

err_exit:
	devglue_exit(devglue);
	class_destroy(devglue->class);
err_clks:
	for(i=0; i<devglue->nr_clks; i++) {
		if (devglue->clks[i].clk) {
			clk_disable_unprepare(devglue->clks[i].clk);
			devm_clk_put(dev, devglue->clks[i].clk);
		}
	}
	devm_kfree(dev, devglue->clks);
err_irqs:
	devm_kfree(dev, devglue->irqs);
err_regs:
	devm_kfree(dev, devglue->regs);
err_work:
	devm_kfree(dev, devglue);

	return ret;
}

/*
 * remove driver
 */
static int devglue_remove(struct platform_device *pdev)
{
	struct device *dev = &pdev->dev;
	struct devglue_uniphier *devglue = platform_get_drvdata(pdev);
	int i;

	if (devglue != _devglue) {
		dev_err(dev, "Unmatch driver call\n");
		return -EINVAL;
	}
	_devglue = NULL;

	/* depopulate driver */
	of_platform_depopulate(dev);

	/* finalize hardware */
	devglue_exit(devglue);

	/* remove device class */
	class_destroy(devglue->class);

	/* disable clock */
	for(i=0; i<devglue->nr_clks; i++) {
		clk_disable_unprepare(devglue->clks[i].clk);
		devm_clk_put(dev, devglue->clks[i].clk);
	}
	devm_kfree(dev, devglue->clks);

	/* remove interrupt */
	devm_kfree(dev, devglue->irqs);

	/* unmap register */
	devm_kfree(dev, devglue->regs);

	devm_kfree(dev, devglue);

	return 0;
}

static const struct of_device_id of_devglue_match[] = {
	{ .compatible = "socionext,devglue", },
	{ /* Sentinel */ }
};
MODULE_DEVICE_TABLE(of, of_devglue_match);

static struct platform_driver devglue_driver = {
	.probe  = devglue_probe,
	.remove = devglue_remove,
	.driver = {
		.name = "devglue",
		.of_match_table	= of_devglue_match,
	},
};

module_platform_driver(devglue_driver);

MODULE_ALIAS("platform:devglue");
MODULE_AUTHOR("Kunihiko Hayashi <hayashi.kunihiko@socionext.com");
MODULE_LICENSE("GPL v2");
MODULE_DESCRIPTION("devglue: device glue driver");
